# -*- coding: utf-8 -*-
"""internet protocol suite

`jspcap.ipsuite` is a collection for protocol constructor
described in Internet Protocol Suite.

"""
# Internet Protocol Suite
from jspcap.ipsuite.link import *
from jspcap.ipsuite.internet import *
from jspcap.ipsuite.transport import *
from jspcap.ipsuite.application import *

# File Specific Headers
from jspcap.ipsuite.pcap import *

# Abstract Base Class
from jspcap.ipsuite.protocol import Protocol as IPSProtocol


__all__ = [
    'IPSHeader', 'IPSFrame'                     # PCAP Headers
]



# -*- coding: utf-8 -*-
"""application layer protocols

`jspcap.protocols.application` is collection of all
protocols in application layer, with detailed
implementation and methods.

"""
# TODO: Implements BGP, DHCP, DNS, FTP, IMAP, IDAP, MQTT, NNTP, NTP, ONC:RPC, POP, RIP, RTP, SIP, SMTP, SNMP, SSH, SSL, TELNET, TLS, XMPP.

# Base Class for Internet Layer
from jspcap.protocols.application.application import Application

# Utility Classes for Protocols
from jspcap.protocols.application.httpv1 import HTTPv1
from jspcap.protocols.application.httpv2 import HTTPv2

# Deprecated / Base Classes
from jspcap.protocols.application.http import HTTP


__all__ = ['HTTPv1', 'HTTPv2']



# -*- coding: utf-8 -*-
"""protocol chain collection

`jspcap.corekit.protochain` contains special protocol
collection class `ProtoChain`.

"""
import numbers

from jspcap.utilities.exceptions import IndexNotFound
from jspcap.utilities.validations import int_check

###############################################################################
# from jspcap.protocols.protocol import Protocol
###############################################################################


__all__ = ['ProtoChain']


class ProtoChain:
    """Protocols chain.

    Properties:
        * alias -- tuple, aliases of protocols in chain
        * tuple -- tuple, name of protocols in chain
        * proto -- tuple, lowercase name of protocols in chain
        * chain -- str, chain of protocols seperated by colons

    Methods:
        * index -- same as `index` function of `tuple` type

    Attributes:
        * __damn__ -- tuple, aliase of protocols in chain
        * __data__ -- tuple, name of protocols in chain

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def alias(self):
        return self.__damn__

    @property
    def tuple(self):
        return self.__data__

    @property
    def proto(self):
        proto = list()
        for name in self.__data__:
            proto.append(str(name).lower().replace('none', 'raw'))
        return tuple(proto)

    @property
    def chain(self):
        return self.__str__()

    ##########################################################################
    # Methods.
    ##########################################################################

    def index(self, name, start=None, stop=None):
        try:
            start = start or 0
            stop = stop or len(self.tuple)

            if isinstance(name, str):
                name = name.lower()
            if isinstance(start, str):
                start = self.index(start)
            if isinstance(stop, str):
                stop = self.index(stop)
            int_check(start, stop)
            return self.proto.index(name, start, stop)
        except ValueError:
            raise IndexNotFound(f"'{name}' not in ProtoChain")

    ##########################################################################
    # Data modules.
    ##########################################################################

    def __init__(self, proto, other=None, alias=None):
        alias = alias or proto
        if other is None:
            self.__data__ = (proto,)
            self.__damn__ = (alias,)
        else:
            self.__data__ = (proto,) + other.tuple
            self.__damn__ = (alias,) + other.alias

    def __repr__(self):
        repr_ = ', '.join(self.proto)
        return f'ProtoChain({repr_})'

    def __str__(self):
        for (i, proto) in enumerate(self.__damn__):
            if proto is None or proto == 'Raw':
                return ':'.join(self.__damn__[:i])
        return ':'.join(self.__damn__)

    def __getitem__(self, key):
        if isinstance(key, slice):
            start = key.start
            stop = key.stop
            step = key.step

            if not isinstance(start, numbers.Number):
                start = self.index(start)
            if not isinstance(stop, numbers.Number):
                stop = self.index(stop)
            int_check(start, stop, step)
            key = slice(start, stop, step)
        elif isinstance(key, numbers.Number):
            key = key
        else:
            key = self.index(key)
        return self.__data__[key]

    def __iter__(self):
        return iter(self.__damn__)

    def __contains__(self, name):
        from jspcap.protocols.protocol import Protocol
        if isinstance(name, type) and issubclass(name, Protocol):
            name = name.__index__()
        if isinstance(name, tuple):
            for item in name:
                flag = (item.lower() in self.proto)
                if flag:    break
            return flag
        if isinstance(name, str):
            name = name.lower()
        return (name in self.proto)



# -*- coding: utf-8 -*-
"""reassembly IP fragments

`jspcap.reassembly.ip` contains `IP_Reassembly` only,
which is the base class for IPv4 and IPv6 reassembly.
The following algorithm implementment is based on IP
reassembly procedure introduced in RFC 791, using
`RCVBT` (fragment receivedbit table). Though another
algorithm is explained in RFC 815, replacing `RCVBT`,
however, this implementment still used the elder one.
And here is the pseudo-code:

Notations:

    FO    - Fragment Offset
    IHL   - Internet Header Length
    MF    - More Fragments flag
    TTL   - Time To Live
    NFB   - Number of Fragment Blocks
    TL    - Total Length
    TDL   - Total Data Length
    BUFID - Buffer Identifier
    RCVBT - Fragment Received Bit Table
    TLB   - Timer Lower Bound

Algorithm:

    DO {
        BUFID <- source|destination|protocol|identification;

        IF (FO = 0 AND MF = 0) {
            IF (buffer with BUFID is allocated) {
                flush all reassembly for this BUFID;
                Submit datagram to next step;
                DONE.
            }
        }

        IF (no buffer with BUFID is allocated) {
            allocate reassembly resources with BUFID;
            TIMER <- TLB;
            TDL <- 0;
            put data from fragment into data buffer with BUFID
                [from octet FO*8 to octet (TL-(IHL*4))+FO*8];
            set RCVBT bits [from FO to FO+((TL-(IHL*4)+7)/8)];
        }

        IF (MF = 0) {
            TDL <- TL-(IHL*4)+(FO*8)
        }

        IF (FO = 0) {
            put header in header buffer
        }

        IF (TDL # 0 AND all RCVBT bits [from 0 to (TDL+7)/8] are set) {
            TL <- TDL+(IHL*4)
            Submit datagram to next step;
            free all reassembly resources for this BUFID;
            DONE.
        }

        TIMER <- MAX(TIMER,TTL);

    } give up until (next fragment or timer expires);

    timer expires: {
        flush all reassembly with this BUFID;
        DONE.
    }

"""
import copy

from jspcap.corekit.infoclass import Info
from jspcap.reassembly.reassembly import Reassembly


__all__ = ['IP_Reassembly']


class IP_Reassembly(Reassembly):
    """Reassembly for IP payload.

    Properties:
        * name -- str, protocol of current packet
        * count -- int, total number of reassembled packets
        * datagram -- tuple, reassembled datagram, which structure may vary
                        according to its protocol
        * protocol -- str, protocol of current reassembly object

    Methods:
        * reassembly -- perform the reassembly procedure
        * submit -- submit reassembled payload
        * fetch -- fetch datagram
        * index -- return datagram index
        * run -- run automatically

    Attributes:
        * _strflg -- bool, stirct mode flag
        * _buffer -- dict, buffer field
        * _dtgram -- tuple, reassembled datagram

    """
    ##########################################################################
    # Methods.
    ##########################################################################

    def reassembly(self, info):
        """Reassembly procedure.

        Positional arguments:
            * info -- Info, info dict of packets to be reassembled

        """
        BUFID = info.bufid  # Buffer Identifier
        FO = info.fo        # Fragment Offset
        IHL = info.ihl      # Internet Header Length
        MF = info.mf        # More Fragments flag
        TL = info.tl        # Total Length

        # when unfragmented (possibly discarded) packet received
        if not FO and not MF:
            if BUFID in self._buffer:
                self._dtgram += self.submit(self._buffer[BUFID])
                del self._buffer[BUFID]
                return

        # initialise buffer with BUFID
        if BUFID not in self._buffer:
            self._buffer[BUFID] = dict(
                TDL = 0,                        # Total Data Length
                RCVBT = bytearray(8191),        # Fragment Received Bit Table
                index = list(),                 # index record
                header = bytearray(),           # header buffer
                datagram = bytearray(65535),    # data buffer
            )

        # append packet index
        self._buffer[BUFID]['index'].append(info.num)

        # put data into data buffer
        start = FO
        stop = TL - IHL + FO
        self._buffer[BUFID]['datagram'][start:stop] = info.payload

        # set RCVBT bits (in 8 octets)
        start = FO // 8
        stop = FO // 8 + (TL - IHL + 7) // 8
        self._buffer[BUFID]['RCVBT'][start:stop] = b'\x01' * (stop - start + 1)

        # get total data length (header excludes)
        if not MF:
            TDL = TL - IHL + FO

        # put header into header buffer
        if not FO:
            self._buffer[BUFID]['header'] = info.header

        # when datagram is reassembled in whole
        start = 0
        stop = (TDL + 7) // 8
        if TDL and all(self._buffer[BUFID]['RCVBT'][start:stop]):
            self._dtgram += self.submit(self._buffer[BUFID], checked=True)
            del self._buffer[BUFID]

    def submit(self, buf, *, checked=False):
        """Submit reassembled payload.

        Positional arguments:
            * buf -- dict, buffer dict of reassembled packets

        Keyword arguments:
            * bufid -- tuple, buffer identifier

        Returns:
            * list -- reassembled packets

        """
        TDL = buf['TDL']
        RCVBT = buf['RCVBT']
        index = buf['index']
        header = buf['header']
        datagram = buf['datagram']

        start = 0
        stop = (TDL + 7) // 8
        flag = checked or (TDL and all(RCVBT[start:stop]))
        # if datagram is not implemented
        if not flag and self._strflg:
            data = list()
            byte = bytearray()
            # extract received payload
            for (bctr, bit) in enumerate(RCVBT):
                if bit: # received bit
                    this = bctr * 8
                    that = this + 8
                    byte += datagram[this:that]
                else:   # missing bit
                    if byte:    # strip empty payload
                        data.append(bytes(byte))
                    byte = bytearray()
            # strip empty packets
            if data or header:
                packet = Info(
                    NotImplemented = True,
                    index = tuple(index),
                    header = header or None,
                    payload = tuple(data) or None,
                )
        # if datagram is reassembled in whole
        else:
            payload = datagram[:TDL]
            packet = Info(
                NotImplemented = False,
                index = tuple(index),
                packet = (bytes(header) + bytes(payload)) or None,
            )
        return [packet]



# -*- coding: utf-8 -*-
"""(dynamic) reverse address resolution protocol

`jspcap.protocols.link.rarp` contains `RARP` only,
which implements extractor for (Dynamic) Reverse Address
Resolution Protocol (RARP/DRARP), whose structure is
described as below.

Octets      Bits        Name                    Discription
  0           0     rarp.htype              Hardware Type
  2          16     rarp.ptype              Protocol Type
  4          32     rarp.hlen               Hardware Address Length
  5          40     rarp.plen               Protocol Address Length
  6          48     rarp.oper               Operation
  8          64     rarp.sha                Sender Hardware Address
  14        112     rarp.spa                Sender Protocol Address
  18        144     rarp.tha                Target Hardware Address
  24        192     rarp.tpa                Target Protocol Address

"""
from jspcap.protocols.link.arp import ARP


__all__ = ['RARP']


class RARP(ARP):
    """This class implements Reverse Address Resolution Protocol.

    Properties:
        * name -- str, name of corresponding protocol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding protocol
        * layer -- str, `Link`
        * length -- int, header length of corresponding protocol
        * protochain -- ProtoChain, protocol chain of current instance
        * src -- tuple(str, str), sender hardware & protocol address
        * dst -- tuple(str, str), target hardware & protocol address
        * type -- tuple(str, str), hardware & protocol type

   Methods:
       * read_arp -- read Address Resolution Protocol

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance
        * _acnm -- str, acronym of corresponding protocol
        * _name -- str, name of corresponding protocol

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor
        * _read_addr_resolve -- resolve MAC address according to protocol
        * _read_proto_resolve -- solve IP address according to protocol

    """
    _acnm = 'RARP'
    _name = 'Reverse Address Resolution Protocol'

    ##########################################################################
    # Data models.
    ##########################################################################

    @classmethod
    def __index__(cls):
        return ('RARP', 'DRARP')



# -*- coding: utf-8 -*-
"""internet protocol

`jspcap.protocols.internet.ipsec` contains `IP` only,
which is a base class for Internet Protocol (IP) protocol
family, eg. IPv4, IPv6, and IPsec.

"""
from jspcap.corekit.infoclass import Info
from jspcap.protocols.internet.internet import Internet
from jspcap.utilities.decorators import seekset


__all__ = ['IP']


class IP(Internet):
    """This class implements all protocols in IP family.

    - Internet Protocol version 4 (IPv4) [RFC 791]
    - Internet Protocol version 6 (IPv6) [RFC 2460]
    - Authentication Header (AH) [RFC 4302]
    - Encapsulating Security Payload (ESP) [RFC 4303]

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Internet`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance
        * src -- str, source IP address
        * dst -- str, destination IP address

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    # source IP address
    @property
    def src(self):
        """Source IP address."""
        return self._info.src

    # destination IP address
    @property
    def dst(self):
        """Destination IP address."""
        return self._info.dst

    ##########################################################################
    # Data models.
    ##########################################################################

    @classmethod
    def __index__(cls):
        return ('IPv4', 'IPv6')



# -*- coding: utf-8 -*-
"""library foundation

`jspcap.foundation` is a collection of fundations for `jspcap`,
including PCAP file extraction tool `Extrator` and application
layer protocol analyser `Analysis`.

"""
from jspcap.foundation.analysis import *
from jspcap.foundation.extraction import *
from jspcap.foundation.traceflow import *


__all__ = ['Analysis', 'Extractor', 'TraceFlow']



# -*- coding: utf-8 -*-
"""raw packet data

`jspcap.protocols.raw` contains `Raw` only, which
implements extractor for unknown protocol, and constructs
a `Protocol` like object.

"""
from jspcap.corekit.infoclass import Info
from jspcap.corekit.protochain import ProtoChain
from jspcap.protocols.protocol import Protocol
from jspcap.utilities.exceptions import UnsupportedCall


__all__ = ['Raw']


class Raw(Protocol):
    """This class implements universal unknown protocol.

    Properties:
        * name -- str, name of corresponding protocol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Methods:
        * read_raw -- read raw packet data

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor
        * _make_protochain -- make ProtoChain instance for corresponding protocol

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    # name of current protocol
    @property
    def name(self):
        """Name of current protocol."""
        return 'Unknown'

    # header length of current protocol
    @property
    def length(self):
        """DEPRECATED"""
        raise UnsupportedCall(f"'{self.__class__.__name__}' object has no attribute 'length'")

    # name of next layer protocol
    @property
    def protocol(self):
        """DEPRECATED"""
        raise UnsupportedCall(f"'{self.__class__.__name__}' object has no attribute 'protocol'")

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_raw(self, length, *, error=None):
        """Read raw packet data."""
        if length is None:
            length = len(self)

        raw = dict(
            packet = self._read_packet(length),
            error = error or None,
        )

        return raw

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, file, length=None, *, error=None, **kwargs):
        self._file = file
        self._info = Info(self.read_raw(length, error=error))
        self._make_protochain()

    def __length_hint__(self):
        pass

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _make_protochain(self):
        """Make ProtoChain instance for corresponding protocol."""
        self._protos = ProtoChain(self.__class__.__name__, None, self.alias)



# -*- coding: utf-8 -*-
"""mobility header

`jspcap.protocols.internet.mh` contains `MH` only,
which implements extractor for Mobility Header (MH),
whose structure is described as below.

+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
| Payload Proto |  Header Len   |   MH Type     |   Reserved    |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|           Checksum            |                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+                               |
|                                                               |
.                                                               .
.                       Message Data                            .
.                                                               .
|                                                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

"""
# TODO: Implements extractor for message data of all MH types.
from jspcap.corekit.infoclass import Info
from jspcap.protocols.protocol import Protocol


__all__ = ['MH']


# Mobility Header Types - for the MH Type field in the Mobility Header
_MOBILITY_TYPE = {
    0 : 'Binding Refresh Request',              # [RFC 6275]
    1 : 'Home Test Init',                       # [RFC 6275]
    2 : 'Care-of Test Init',                    # [RFC 6275]
    3 : 'Home Test',                            # [RFC 6275]
    4 : 'Care-of Test',                         # [RFC 6275]
    5 : 'Binding Update',                       # [RFC 6275]
    6 : 'Binding Acknowledgement',              # [RFC 6275]
    7 : 'Binding Error',                        # [RFC 6275]
    8 : 'Fast Binding Update',                  # [RFC 5568]
    9 : 'Fast Binding Acknowledgment',          # [RFC 5568]
   10 : 'Fast Neighbor Advertisement',          # [RFC 5568] DEPRECATED
   11 : 'Experimental Mobility Header',         # [RFC 5096]
   12 : 'Home Agent Switch Message',            # [RFC 5142]
   13 : 'Heartbeat Message',                    # [RFC 5847]
   14 : 'Handover Initiate Message',            # [RFC 5568]
   15 : 'Handover Acknowledge Message',         # [RFC 5568]
   16 : 'Binding Revocation Message',           # [RFC 5846]
   17 : 'Localized Routing Initiation',         # [RFC 6705]
   18 : 'Localized Routing Acknowledgment',     # [RFC 6705]
   19 : 'Update Notification',                  # [RFC 7077]
   20 : 'Update Notification Acknowledgement',  # [RFC 7077]
   21 : 'Flow Binding Message',                 # [RFC 7109]
   22 : 'Subscription Query',                   # [RFC 7161]
   23 : 'Subscription Response',                # [RFC 7161]
}


class MH(Protocol):
    """This class implements Mobility Header.

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Internet`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Methods:
        * read_mh -- read Mobility Header (MH)

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of current protocol."""
        return 'Mobility Header'

    @property
    def length(self):
        """Header length of current protocol."""
        return self._info.length

    @property
    def protocol(self):
        """Name of next layer protocol."""
        return self._info.next

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_mh(self, length, extension):
        """Read Mobility Header.

        Structure of MH header [RFC 6275]:
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            | Payload Proto |  Header Len   |   MH Type     |   Reserved    |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |           Checksum            |                               |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+                               |
            |                                                               |
            .                                                               .
            .                       Message Data                            .
            .                                                               .
            |                                                               |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

            Octets      Bits        Name                    Discription
              0           0     mh.next                 Next Header
              1           8     mh.length               Header Length
              2          16     mh.type                 Mobility Header Type
              3          24     -                       Reserved
              4          32     mh.chksum               Checksum
              6          48     mh.data                 Message Data

        """
        if length is None:
            length = len(self)

        _next = self._read_protos(1)
        _hlen = self._read_unpack(1)
        _type = self._read_unpack(1)
        _temp = self._read_fileng(1)
        _csum = self._read_fileng(2)
        _data = self._read_fileng(_hlen - 5)

        mh = dict(
            next = _next,
            length = _hlen + 1,
            type = _MOBILITY_TYPE.get(_type, 'Unassigned'),
            chksum = _csum,
            data = _data,
        )

        length -= mh['length']
        mh['packet'] = self._read_packet(header=mh['length'], payload=length)

        if extension:
            self._protos = None
            return mh
        return self._decode_next_layer(mh, _next, length)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, *, extension=False, **kwargs):
        self._file = _file
        self._info = Info(self.read_mh(length, extension))

    def __length_hint__(self):
        return 6



# -*- coding: utf-8 -*-
"""user interface

`jspcap.interface` defines several user-oriented
interfaces, variables, and etc. These interfaces are
designed to help and simplify the usage of `jspcap`.

"""
from jspcap.foundation.analysis import Analysis
from jspcap.foundation.extraction import Extractor
from jspcap.foundation.traceflow import TraceFlow
from jspcap.protocols.protocol import Protocol
from jspcap.reassembly.ipv4 import IPv4_Reassembly
from jspcap.reassembly.ipv6 import IPv6_Reassembly
from jspcap.reassembly.tcp import TCP_Reassembly
from jspcap.utilities.exceptions import FormatError
from jspcap.utilities.validations import bool_check, int_check, io_check, str_check


__all__ = [
    'extract', 'analyse', 'reassemble', 'trace',            # functions
    'TREE', 'JSON', 'PLIST', 'PCAP',                        # macros
]


# output file formats
TREE  = 'tree'
JSON  = 'json'
PLIST = 'plist'
PCAP  = 'pcap'


def extract(*, fin=None, fout=None, format=None,                            # basic settings
                auto=True, extension=True, store=True,                      # internal settings
                files=False, nofile=False, verbose=False,                   # output settings
                engine=None, layer=None, protocol=None,                     # extraction settings
                ip=False, ipv4=False, ipv6=False, tcp=False, strict=False,  # reassembly settings
                trace=False, trace_fout=None, trace_format=None):           # trace settings
    """Extract a PCAP file.

    Keyword arguments:
        * fin  -- str, file name to be read; if file not exist, raise an error
        * fout -- str, file name to be written
        * format  -- str, file format of output
                        <keyword> 'plist' / 'json' / 'tree' / 'html'

        * auto -- bool, if automatically run till EOF (default is True)
                        <keyword> True / False
        * extension -- bool, if check and append axtensions to output file (default is True)
                        <keyword> True / False
        * store -- bool, if store extracted packet info (default is True)
                        <keyword> True / False

        * files -- bool, if split each frame into different files (default is False)
                        <keyword> True / False
        * nofile -- bool, if no output file is to be dumped (default is False)
                        <keyword> True / False
        * verbose -- bool, if print verbose output information (default is False)
                        <keyword> True / False

        * engine -- str, extraction engine to be used
                        <keyword> 'default | jspcap'
        * layer -- str, extract til which layer
                        <keyword> 'Link' / 'Internet' / 'Transport' / 'Application'
        * protocol -- str, extract til which protocol
                        <keyword> available protocol name

        * ip -- bool, if record data for IPv4 & IPv6 reassembly (default is False)
                        <keyword> True / False
        * ipv4 -- bool, if perform IPv4 reassembly (default is False)
                        <keyword> True / False
        * ipv6 -- bool, if perform IPv6 reassembly (default is False)
                        <keyword> True / False
        * tcp -- bool, if perform TCP reassembly (default is False)
                        <keyword> True / False
        * strict -- bool, if set strict flag for reassembly (default is False)
                        <keyword> True / False

        * trace -- bool, if trace TCP traffic flows (default is False)
                        <keyword> True / False
        * trace_fout -- str, path name for flow tracer if necessary
        * trace_format -- str, output file format of flow tracer
                        <keyword> 'plist' / 'json' / 'tree' / 'html' / 'pcap'

    Returns:
        * Extractor -- an Extractor object form `jspcap.extractor`

    """
    if isinstance(layer, type) and issubclass(layer, Protocol):
        layer = layer.__layer__
    if isinstance(protocol, type) and issubclass(protocol, Protocol):
        protocol = protocol.__index__()

    str_check(fin or '', fout or '', format or '',
                trace_fout or '', trace_format or '',
                engine or '', layer or '', *(protocol or ''))
    bool_check(files, nofile, verbose, auto, extension, store, 
                ip, ipv4, ipv6, tcp, strict, trace)

    return Extractor(fin=fin, fout=fout, format=format,
                        store=store, files=files, nofile=nofile,
                        auto=auto, verbose=verbose, extension=extension,
                        engine=engine, layer=layer, protocol=protocol,
                        ip=ip, ipv4=ipv4, ipv6=ipv6, tcp=tcp, strict=strict,
                        trace=trace, trace_fout=trace_fout, trace_format=trace_format)


def analyse(*, file, length=None):
    """Analyse application layer packets.

    Keyword arguments:
        * file -- file-like object, packet to be analysed
        * length -- int, length of the analysing packet

    Returns:
        * Analysis -- an Analysis object from `jspcap.analyser`

    """
    io_check(file)
    int_check(length)

    return Analysis.analyse(file, length)


def reassemble(*, protocol, strict=False):
    """Reassemble fragmented datagrams.

    Keyword arguments:
        * protocol -- str, protocol to be reassembled
        * strict -- bool, if return all datagrams (including those not implemented) when submit (default is False)
                        <keyword> True / False

    Returns:
        * [if protocol is IPv4] IPv4_Reassembly -- a Reassembly object from `jspcap.reassembly`
        * [if protocol is IPv6] IPv6_Reassembly -- a Reassembly object from `jspcap.reassembly`
        * [if protocol is TCP] TCP_Reassembly -- a Reassembly object from `jspcap.reassembly`

    """
    if isinstance(protocol, type) and issubclass(protocol, Protocol):
        protocol = protocol.__index__()

    str_check(protocol)
    bool_check(strict)

    if protocol == 'IPv4':
        return IPv4_Reassembly(strict=strict)
    elif protocol == 'IPv6':
        return IPv6_Reassembly(strict=strict)
    elif protocol == 'TCP':
        return TCP_Reassembly(strict=strict)
    else:
        raise FormatError(f'Unsupported reassembly protocol: {protocol}')


def trace(*, fout=None, format=None):
    """Trace TCP flows.

    """
    str_check(fout or '', format or '')
    return TraceFlow(fout, format)



# -*- coding: utf-8 -*-
"""layer two tunnelling protocol

`jspcap.protocols.link.l2tp` contains `L2TP` only,
which implements extractor for Layer Two Tunnelling Protocol
(L2TP), whose structure is described as below.

 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|T|L|x|x|S|x|O|P|x|x|x|x|  Ver  |          Length (opt)         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|           Tunnel ID           |           Session ID          |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|             Ns (opt)          |             Nr (opt)          |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|      Offset Size (opt)        |    Offset pad... (opt)
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

"""
from jspcap.corekit.infoclass import Info
from jspcap.protocols.link.link import Link


__all__ = ['L2TP']


class L2TP(Link):
    """This class implements Layer Two Tunnelling Protocol.

    Properties:
        * name -- str, name of corresponding protocol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding protocol
        * layer -- str, `Link`
        * length -- int, header length of corresponding protocol
        * type -- str, L2TP packet type
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance
        * type -- str, L2TP type

    Methods:
        * read_l2tp -- read Layer Two Tunnelling Protocol

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of current protocol."""
        return 'Layer 2 Tunnelling Protocol'

    @property
    def length(self):
        """Header length of current protocol."""
        return self._info.hdr_len

    @property
    def type(self):
        """L2TP type."""
        return self._info.flags.type

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_l2tp(self, length):
        """Read Layer Two Tunnelling Protocol.

        Structure of L2TP header [RFC 2661]:
             0                   1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |T|L|x|x|S|x|O|P|x|x|x|x|  Ver  |          Length (opt)         |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |           Tunnel ID           |           Session ID          |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |             Ns (opt)          |             Nr (opt)          |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |      Offset Size (opt)        |    Offset pad... (opt)
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

            Octets      Bits        Name                    Discription
              0           0     l2tp.flags              Flags and Version Info
              0           0     l2tp.flags.type         Type (0/1)
              0           1     l2tp.flags.len          Length
              0           2     -                       Reserved (must be zero)
              0           4     l2tp.flags.seq          Sequence
              0           5     -                       Reserved (must be zero)
              0           6     l2tp.flags.offset       Offset
              0           7     l2tp.flags.prio         Priority
              1           8     -                       Reserved (must be zero)
              1          12     l2tp.ver                Version (2)
              2          16     l2tp.length             Length (optional by len)
              4          32     l2tp.tunnelid           Tunnel ID
              6          48     l2tp.sessionid          Session ID
              8          64     l2tp.ns                 Sequence Number (optional by seq)
              10         80     l2tp.nr                 Next Sequence Number (optional by seq)
              12         96     l2tp.offset             Offset Size (optional by offset)

        """
        if length is None:
            length = len(self)

        _flag = self._read_binary(1)
        _vers = self._read_fileng(1).hex()[1]
        _hlen = self._read_unpack(2) if int(_flag[1]) else None
        _tnnl = self._read_unpack(2)
        _sssn = self._read_unpack(2)
        _nseq = self._read_unpack(2) if int(_flag[4]) else None
        _nrec = self._read_unpack(2) if int(_flag[4]) else None
        _size = self._read_unpack(2) if int(_flag[6]) else 0

        l2tp = dict(
            flags = dict(
                type = 'Control' if int(_flag[0]) else 'Data',
                len = True if int(_flag[1]) else False,
                seq = True if int(_flag[4]) else False,
                offset = True if int(_flag[6]) else False,
                prio = True if int(_flag[7]) else False,
            ),
            ver = int(_vers, base=16),
            length = _hlen,
            tunnelid = _tnnl,
            sessionid = _sssn,
            ns = _nseq,
            nr = _nrec,
            offset = 8*_size or None,
        )

        hdr_len = _hlen or (6 + 2*(int(_flag[1]) + 2*int(_flag[4]) + int(_flag[6])))
        l2tp['hdr_len'] = hdr_len + _size * 8
        # if _size:
            # l2tp['padding'] = self._read_fileng(_size * 8)

        length -= l2tp['hdr_len']
        l2tp['packet'] = self._read_packet(header=l2tp['hdr_len'], payload=length)

        return self._decode_next_layer(l2tp, length)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, **kwargs):
        self._file = _file
        self._info = Info(self.read_l2tp(length))

    def __length_hint__(self):
        return 16



# -*- coding: utf-8 -*-
"""reassembly fragmented packets

`jspcap.reassembly.reassembly` contains `Reassembly`
only, which is an abstract base class for all reassembly
classes, bases on algorithms described in
[`RFC 815`](https://tools.ietf.org/html/rfc815),
implements datagram reassembly of IP and TCP packets.

"""
import abc
import copy

from jspcap.corekit.infoclass import Info
from jspcap.utilities.validations import frag_check, int_check


__all__ = ['Reassembly']


ABCMeta = abc.ABCMeta
abstractmethod = abc.abstractmethod
abstractproperty = abc.abstractproperty


class Reassembly(object):
    """Base class for reassembly procedure.

    Properties:
        * name -- str, name of current protocol
        * count -- int, total number of reassembled packets
        * datagram -- tuple, reassembled datagram, which structure may vary
                        according to its protocol
        * protocol -- str, protocol of current reassembly object

    Methods:
        * reassembly -- perform the reassembly procedure
        * submit -- submit reassembled payload
        * fetch -- fetch datagram
        * index -- return datagram index
        * run -- run automatically

    Attributes:
        * _strflg -- bool, stirct mode flag
        * _newflg -- bool, if new packets reassembled flag
        * _buffer -- dict, buffer field
        * _dtgram -- list, reassembled datagram

    """
    __metaclass__ = ABCMeta

    ##########################################################################
    # Properties.
    ##########################################################################

    # protocol of current packet
    @abstractproperty
    def name(self):
        """Protocol of current packet."""
        pass

    # total number of reassembled packets
    @property
    def count(self):
        """Total number of reassembled packets."""
        return len(self.fetch())

    # reassembled datagram
    @property
    def datagram(self):
        """Reassembled datagram."""
        return self.fetch()

    @abstractproperty
    def protocol(self):
        """Protocol of current reassembly object."""
        pass

    ##########################################################################
    # Methods.
    ##########################################################################

    # reassembly procedure
    @abstractmethod
    def reassembly(self, info):
        """Reassembly procedure.

        Positional arguments:
            * info - Info, info dict of packets to be reassembled

        Returns:
            * NotImplemented

        """
        pass

    # submit reassembled payload
    @abstractmethod
    def submit(self, buf, **kwargs):
        """Submit reassembled payload.

        Positional arguments:
            * buf -- dict, buffer dict of reassembled packets

        Returns:
            * NotImplemented

        """
        pass

    # fetch datagram
    def fetch(self):
        """Fetch datagram."""
        if self._newflg:
            self._newflg = False
            temp_dtgram = copy.deepcopy(self._dtgram)
            for (bufid, buffer) in self._buffer.items():
                temp_dtgram += self.submit(buffer, bufid=bufid)
            return tuple(temp_dtgram)
        return tuple(self._dtgram)

    # return datagram index
    def index(self, pkt_num):
        """Return datagram index."""
        int_check(pkt_num)
        for counter, datagram in enumerate(self.datagram):
            if pkt_num in datagram.index:
                return counter
        return None

    # run automatically
    def run(self, packets):
        """Run automatically.

        Positional arguments:
            * packets -- list<dict>, list of packet dicts to be reassembled

        """
        for packet in packets:
            frag_check(packet, protocol=self.protocol)
            info = Info(packet)
            self.reassembly(info)
        self._newflg = True

    ##########################################################################
    # Data models.
    ##########################################################################

    # Not hashable
    __hash__ = None

    def __init__(self, *, strict=False):
        """Initialise packet reassembly.

        Keyword arguments:
            * strict -- bool, if return all datagrams (including those not
                        implemented) when submit (default is False)
                            <keyword> True / False

        """
        self._newflg = False    # new packets reassembled
        self._strflg = strict   # stirct mode flag
        self._buffer = dict()   # buffer field
        self._dtgram = list()   # reassembled datagram

    def __call__(self, packet):
        """Call packet reassembly.

        Positional arguments:
            * packet -- dict, packet dict to be reassembled
                        (detailed format described in corresponding protocol)

        """
        frag_check(packet, protocol=self.protocol)
        info = Info(packet)
        self.reassembly(info)
        self._newflg = True



# -*- coding: utf-8 -*-
"""info class

`jspcap.corekit.infoclass` contains dict-like class
`Info` only, which is originally designed to work alike
`dataclasses.dataclass` in Python 3.7 and later versions.

"""
import copy

from jspcap.utilities.exceptions import UnsupportedCall
from jspcap.utilities.validations import dict_check


__all__ = ['Info']


class Info(dict):
    """Turn dictionaries into object-like instances.

    Methods:
        * infotodict -- reverse Info object into dict type

    Notes:
        * Info objects inherit from `dict` type
        * Info objects are iterable, and support all functions as `dict`
        * Info objects are one-time-modeling, thus cannot set or delete
            attributes after initialisation

    """
    def __new__(cls, dict_=None, **kwargs):
        def __read__(dict_):
            __dict__ = dict()
            for (key, value) in dict_.items():
                if isinstance(value, dict):
                    __dict__[key] = Info(value)
                else:
                    # if isinstance(key, str):
                    #     key = re.sub('\W', '_', key)
                    __dict__[key] = value
            return __dict__

        self = super().__new__(cls)
        if dict_ is not None:
            if isinstance(dict_, Info):
                self = copy.deepcopy(dict_)
            else:
                dict_check(dict_)
                self.__dict__.update(__read__(dict_))

        self.__dict__.update(__read__(kwargs))
        return self

    def __repr__(self):
        temp = list()
        for (key, value) in self.__dict__.items():
            temp.append(f'{key}={value}')
        args = ', '.join(temp)
        return f'Info({args})'

    __str__ = __repr__

    def __iter__(self):
        return iter(self.__dict__)

    def __getitem__(self, key):
        return self.__dict__[key]

    def __contains__(self, name):
        return (name in self.__dict__)

    def __setattr__(self, name, value):
        raise UnsupportedCall("can't set attribute")

    def __delattr__(self, name):
        raise UnsupportedCall("can't delete attribute")

    def infotodict(self):
        dict_ = dict()
        for (key, value) in self.__dict__.items():
            if isinstance(value, Info):
                dict_[key] = value.infotodict()
            elif isinstance(value, (tuple, list)):
                temp = list()
                for item in value:
                    if isinstance(item, Info):
                        temp.append(item.infotodict())
                    else:
                        temp.append(item)
                dict_[key] = value.__class__(temp)
            else:
                dict_[key] = value
        return dict_



# -*- coding: utf-8 -*-
"""transport layer protocols

`jspcap.protocols.transport` is collection of all protocols
in transport layer, with detailed implementation and
methods.

"""
# TODO: Implements DCCP, RSVP, STCP.

# Base Class for Transport Layer
from jspcap.protocols.transport.transport import Transport

# Utility Classes for Protocols
from jspcap.protocols.transport.tcp import TCP
from jspcap.protocols.transport.udp import UDP

# Transport Layer Protocol Numbers
from jspcap.protocols.transport.transport import TP_PROTO


__all__ = [
    'TP_PROTO',     # Protocol Numbers
    'TCP', 'UDP',   # Transport Layer Protocols
]



# -*- coding: utf-8 -*-
"""protocol family

`jspcap.protocols` is collection of all protocol families,
with detailed implementation and methods.

"""
# TODO: Implement specified classes for MAC and IP addresses.

# Base Class for Protocols
from jspcap.protocols.protocol import Protocol

# Utility Classes for Protocols
from jspcap.protocols.pcap import *
from jspcap.protocols.raw import Raw

# Protocols & Macros
from jspcap.protocols.link import *
from jspcap.protocols.internet import *
from jspcap.protocols.transport import *
from jspcap.protocols.application import *

# Deprecated / Base Protocols
from jspcap.protocols.internet.ip import IP
from jspcap.protocols.internet.ipsec import IPsec
from jspcap.protocols.application.http import HTTP


__all__ = [
    'LINKTYPE', 'ETHERTYPE', 'TP_PROTO',                # Protocol Numbers
    'Header', 'Frame',                                  # PCAP Headers
    'Raw',                                              # Raw Packet
    'ARP', 'DRARP', 'Ethernet', 'InARP', 'L2TP', 'OSPF', 'RARP', 'VLAN',
                                                        # Link Layer
    'AH', 'IP', 'IPsec', 'IPv4', 'IPv6', 'IPX',         # Internet Layer
    'HIP', 'HOPOPT', 'IPv6_Frag', 'IPv6_Opts', 'IPv6_Route', 'MH',
                                                        # IPv6 Extension Header
    'TCP', 'UDP',                                       # Transport Layer
    'HTTP',                                             # Application Layer
]



# -*- coding: utf-8 -*-
"""

"""
import copy
import pathlib
import warnings

###############################################################################
# from jsformat import PLIST, JSON, Tree, JavaScript, XML
###############################################################################

from jspcap.corekit.infoclass import Info
from jspcap.utilities.exceptions import stacklevel, FileExists
from jspcap.utilities.warnings import FormatWarning, FileWarning
from jspcap.utilities.validations import pkt_check

###############################################################################
# from jspcap.dumpkit import PCAP, NotImplementedIO
###############################################################################


class TraceFlow:
    """Trace TCP flows.

    Properties:
        * index -- tuple<Info>, index table for traced flows

    Methods:
        * make_fout -- make root path for output
        * dump -- dump frame to output files
        * trace -- trace packets

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def index(self):
        if self._newflg:
            return self.submit()
        return tuple(self._stream)

    ##########################################################################
    # Methods.
    ##########################################################################

    @staticmethod
    def make_fout(fout='./tmp', fmt='pcap'):
        """Make root path for output.

        Positional arguments:
            * fout -- str, root path for output
            * fmt -- str, output format

        Returns:
            * output -- dumper of specified format

        """
        if fmt == 'pcap':
            from jspcap.dumpkit import PCAP as output       # output PCAP file 
        elif fmt == 'plist':
            from jsformat import PLIST as output            # output PLIST file
        elif fmt == 'json':
            from jsformat import JSON as output             # output JSON file
        elif fmt == 'tree':
            from jsformat import Tree as output             # output treeview text file
            fmt = 'txt'
        elif fmt == 'html':
            from jsformat import JavaScript as output       # output JavaScript file
            fmt = 'js'
        elif fmt == 'xml':
            from jsformat import XML as output              # output XML file
        else:
            from jspcap.dumpkit import NotImplementedIO as output
                                                            # no output file
            if fmt is not None:
                warnings.warn(f'Unsupported output format: {fmt}; '
                                'disabled file output feature',
                                FormatWarning, stacklevel=stacklevel())

        try:
            pathlib.Path(fout).mkdir(parents=True, exist_ok=True)
        except FileExistsError as error:
            if fmt is None:
                warnings.warn(str(error), FileWarning, stacklevel=stacklevel())
            else:
                raise FileExists(str(error)) from None

        return output, fmt

    def dump(self, packet):
        """Dump frame to output files.

        Positional arguments:
            * packet -- dict, a flow packet
                |-- (str) protocol -- data link type from global header
                |-- (int) index -- frame number
                |-- (Info) frame -- extracted frame info
                |-- (bool) syn -- TCP synchronise (SYN) flag
                |-- (bool) fin -- TCP finish (FIN) flag
                |-- (str) src -- source IP
                |-- (int) srcport -- TCP source port
                |-- (str) dst -- destination IP
                |-- (int) dstport -- TCP destination port
                |-- (numbers.Real) timestamp -- frame timestamp

        """
        # fetch flow label
        output = self.trace(packet, _check=False, _output=True)

        # dump files
        output(packet['frame'], name=f"Frame {packet['index']}")

    def trace(self, packet, *, _check=True, _output=False):
        """Trace packets.

        """
        self._newflg = True
        if _check:
            pkt_check(packet)
        info = Info(packet)

        BUFID = tuple(sorted([info.src, str(info.srcport), info.dst, str(info.dstport)]))
                            # Buffer Identifier
        SYN = info.syn      # Synchronise Flag (Establishment)
        FIN = info.fin      # Finish Flag (Termination)

        # # when SYN is set, reset buffer of this seesion
        # if SYN and BUFID in self._buffer:
        #     temp = self._buffer.pop(BUFID)
        #     temp['fpout'] = (self._fproot, self._fdpext)
        #     temp['index'] = tuple(temp['index'])
        #     self._stream.append(Info(temp))

        # initialise buffer with BUFID
        if BUFID not in self._buffer:
            label = f'{info.src}_{info.srcport}-{info.dst}_{info.dstport}-{info.timestamp}'.replace('.', '_')
            self._buffer[BUFID] = dict(
                fpout = self._foutio(f'{self._fproot}/{label}.{self._fdpext}', protocol=info.protocol),
                index = list(),
                label = label,
            )

        # trace frame record
        self._buffer[BUFID]['index'].append(info.index)
        fpout = self._buffer[BUFID]['fpout']
        label = self._buffer[BUFID]['label']

        # when FIN is set, submit buffer of this session
        if FIN:
            buf = self._buffer.pop(BUFID)
            fpout, label = buf['fpout'], buf['label']
            buf['fpout'] = f'{self._fproot}/{label}.{self._fdpext}'
            buf['index'] = tuple(buf['index'])
            self._stream.append(Info(buf))

        # return label or output object
        return fpout if _output else label

    def submit(self):
        """Submit traced TCP flows."""
        self._newflg = False
        ret = list()
        for buf in self._buffer.values():
            buf = copy.deepcopy(buf)
            buf['fpout'] = f"{self._fproot}/{buf['label']}.{self._fdpext}"
            buf['index'] = tuple(buf['index'])
            ret.append(Info(buf))
        ret += self._stream
        return tuple(ret)

    ##########################################################################
    # Data models.
    ##########################################################################

    # Not hashable
    __hash__ = None

    def __init__(self, *, fout=None, format=None):
        """Initialise instance.

        """
        self._newflg = False    # new packet flag
        self._fproot = fout     # output root path
        self._buffer = dict()   # buffer field
        self._stream = list()   # stream index
        self._foutio, self._fdpext \
                    = self.make_fout(fout, format)
                                # dump I/O object

    def __call__(self, packet):
        """Dump frame to output files."""
        return self.dump(packet)



# -*- coding: utf-8 -*-
"""user defined warnings

"""
import warnings


__all__ = [
    'BaseWarning',                                                  # Warning
    'FormatWarning', 'EngineWarning',                               # ImportWarning
    'FileWarning', 'LayerWarning', 'ProtocolWarning', 'AttributeWarning',
                                                                    # RuntimeWarning
]


##############################################################################
# BaseWarning (abc of warnings) session.
##############################################################################


class BaseWarning(Warning):
    """Base warning class of all kinds."""
    def __init__(self, *args, **kwargs):
        warnings.simplefilter('default')
        super().__init__(*args, **kwargs)


##############################################################################
# ImportWarning session.
##############################################################################


class FormatWarning(BaseWarning, ImportWarning):
    """Warning on unknown format(s)."""
    pass


class EngineWarning(BaseWarning, ImportWarning):
    """Unsupported extraction engine."""
    pass


##############################################################################
# RuntimeWarning session.
##############################################################################


class FileWarning(BaseWarning, RuntimeWarning):
    """Warning on file(s)."""
    pass


class LayerWarning(BaseWarning, RuntimeWarning):
    """Unrecognised layer."""
    pass


class ProtocolWarning(BaseWarning, RuntimeWarning):
    """Unrecognised protocol."""
    pass


class AttributeWarning(BaseWarning, RuntimeWarning):
    """Unsupported attribute."""
    pass



# -*- coding: utf-8 -*-
"""frame header

`jspcap.protocols.pcap.frame` contains `Frame` only,
which implements extractor for frame headers of PCAP,
whose structure is described as below.

typedef struct pcaprec_hdr_s {
    guint32 ts_sec;     /* timestamp seconds */
    guint32 ts_usec;    /* timestamp microseconds */
    guint32 incl_len;   /* number of octets of packet saved in file */
    guint32 orig_len;   /* actual length of packet */
} pcaprec_hdr_t;

"""
import datetime
import io
import os

from jspcap.corekit.infoclass import Info
from jspcap.corekit.protochain import ProtoChain
from jspcap.protocols.protocol import Protocol
from jspcap.utilities.decorators import beholder
from jspcap.utilities.exceptions import ProtocolNotFound, ProtocolUnbound

###############################################################################
# from jspcap.protocols.link import Ethernet
# from jspcap.protocols.internet import IPv4
# from jspcap.protocols.internet import IPv6
# from jspcap.protocols.raw import Raw
###############################################################################


__all__ = ['Frame']


class Frame(Protocol):
    """Per packet frame header extractor.

    Properties:
        * name -- str, name of corresponding protocol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding protocol
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current frame

    Methods:
        * index -- call `ProtoChain.index`
        * read_frame -- read each block after global header

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of corresponding protocol."""
        return f'Frame {self._fnum}'

    @property
    def length(self):
        """Header length of corresponding protocol."""
        return 16

    ##########################################################################
    # Methods.
    ##########################################################################

    def index(self, name):
        return self._proto.index(name)

    def read_frame(self):
        """Read each block after global header.

        Structure of record/package header (C):
            typedef struct pcaprec_hdr_s {
                guint32 ts_sec;     /* timestamp seconds */
                guint32 ts_usec;    /* timestamp microseconds */
                guint32 incl_len;   /* number of octets of packet saved in file */
                guint32 orig_len;   /* actual length of packet */
            } pcaprec_hdr_t;

        """
        # _scur = self._file.tell()
        _temp = self._read_unpack(4, lilendian=True, quiet=True)
        if _temp is None:
            raise EOFError

        _tsss = _temp
        _tsus = self._read_unpack(4, lilendian=True)
        _ilen = self._read_unpack(4, lilendian=True)
        _olen = self._read_unpack(4, lilendian=True)

        _epch = float(f'{_tsss}.{_tsus}')
        _time = datetime.datetime.fromtimestamp(_epch)

        frame = dict(
            frame_info = dict(
                ts_sec = _tsss,
                ts_usec = _tsus,
                incl_len = _ilen,
                orig_len = _olen,
            ),
            time = _time,
            number = self._fnum,
            time_epoch = _epch,
            len = _ilen,
            cap_len = _olen,
        )

        # make BytesIO from frame package data
        length = frame['len']
        self._file = io.BytesIO(self._file.read(length))
        frame['packet'] = self._read_packet(header=0, payload=length, discard=True)

        return self._decode_next_layer(frame, length)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, file, *, num, proto, **kwrags):
        self._fnum = num
        self._file = file
        self._prot = proto
        self._info = Info(self.read_frame())

    def __length_hint__(self):
        return 16

    def __getitem__(self, key):
        if isinstance(key, type) and issubclass(key, Protocol):
            key = key.__index__()

        # if requests attributes in info dict
        if key in self._info:
            return self._info[key]

        def _getitem_from_ProtoChain(key):
            proto = self._protos[key]
            if not proto:
                raise ProtocolNotFound('ProtoChain index out of range')
            elif isinstance(proto, tuple):
                if len(proto) > 1:  # if it's a slice with step & stop
                    raise ProtocolUnbound('frame slice unbound')
                else:
                    start = proto[0]
            else:
                start = self._protos.index(proto)
            return start

        # fetch slice start point from ProtoChain
        if not isinstance(key, tuple):
            key = (key,)
        start = None
        for item in key:
            try:
                start = _getitem_from_ProtoChain(item)
            except ProtocolNotFound:
                continue
            else:
                break
        if start is None:
            raise IndexNotFound(f"'{key}' not in Frame")

        # make return Info item
        dict_ = self._info.infotodict()
        for (level, proto) in enumerate(self._protos):
            proto = proto or 'raw'
            dict_ = dict_[proto.lower()]
            if level >= start:
                return Info(dict_)
        return Info(dict_)

    def __index__(self=None):
        if self is None:    return 'Frame'
        if getattr(self, '_fnum', NotImplemented) is NotImplemented:
            return self.__class__.__name__
        return self._fnum

    def __contains__(self, name):
        if isinstance(name, type) and issubclass(name, Protocol):
            name = name.__index__()
        if isinstance(name, tuple):
            for item in name:
                flag = (item in self._protos)
                if flag:    break
            return flag
        return ((name in self._info) or (name in self._protos))

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _decode_next_layer(self, dict_, length=None):
        """Decode next layer protocol.

        Positional arguments:
            dict_ -- dict, info buffer
            proto -- str, next layer protocol name
            length -- int, valid (not padding) length

        Returns:
            * dict -- current protocol with packet extracted

        """
        seek_cur = self._file.tell()
        try:
            flag, info, chain, alias = self._import_next_layer(self._prot, length)
        except Exception as error:
            dict_['error'] = str(error)
            self._file.seek(seek_cur, os.SEEK_SET)
            flag, info, chain, alias = beholder(self._import_next_layer)(self, self._prot, length, error=True)

        # make next layer protocol name
        if flag:
            proto, name = str(self._prot or 'Raw').lower(), self._prot
        else:
            proto, name = 'raw', 'Raw'

        # write info and protocol chain into dict
        self._protos = ProtoChain(name, chain, alias)
        dict_[proto] = info
        dict_['protocols'] = self._protos.chain
        return dict_

    def _import_next_layer(self, proto, length, error=False):
        """Import next layer extractor.

        Positional arguments:
            * proto -- str, next layer protocol name
            * length -- int, valid (not padding) length

        Keyword arguments:
            * error -- bool, if function call on error

        Returns:
            * bool -- flag if extraction of next layer succeeded
            * Info -- info of next layer
            * ProtoChain -- protocol chain of next layer
            * str -- alias of next layer

        Protocols:
            * Ethernet (data link layer)
            * IPv4 (internet layer)
            * IPv6 (internet layer)

        """
        if proto == 'Ethernet':
            from jspcap.protocols.link import Ethernet as Protocol
        elif proto == 'IPv4':
            from jspcap.protocols.internet import IPv4 as Protocol
        elif proto == 'IPv6':
            from jspcap.protocols.internet import IPv6 as Protocol
        else:
            from jspcap.protocols.raw import Raw as Protocol
        next_ = Protocol(self._file, length, error=error,
                            layer=self._exlayer, protocol=self._exproto)
        return True, next_.info, next_.protochain, next_.alias



# -*- coding: utf-8 -*-
"""PCAP file headers

`jspcap.ipsuite.pcap` contains header constructors for
PCAP files, including global header and frame header.

"""
from jspcap.ipsuite.pcap.header import Header as IPSHeader
from jspcap.ipsuite.pcap.frame import Frame as IPSFrame


__all__ = ['IPSHeader', 'IPSFrame']



# -*- coding: utf-8 -*-
"""internetwork packet exchange

`jspcap.protocols.internet.ipx` contains `IPX` only,
which implements extractor for Internetwork Packet
Exchange (IPX), whose structure is described as below.

Octets      Bits        Name                    Discription
  0           0     ipx.cksum               Checksum
  2          16     ipx.len                 Packet Length (header includes)
  4          32     ipx.count               Transport Control (hop count)
  5          40     ipx.type                Packet Type
  6          48     ipx.dst                 Destination Address
  18        144     ipx.src                 Source Address

"""
from jspcap.corekit.infoclass import Info
from jspcap.protocols.internet.internet import Internet


__all__ = ['IPX']


# IPX Packet Types
TYPE = {
    0 : 'Unknown',
    1 : 'RIP',          # Routing Information Protocol [RFC 1582][RFC 2091]
    2 : 'Echo Packet',
    3 : 'Error Packet',
    4 : 'PEP',          # Packet Exchange Protocol, used for SAP (Service Advertising Protocol)
    5 : 'SPX',          # Sequenced Packet Exchange
    6 : 'NCP',          # NetWare Core Protocol
}

# Socket Types
SOCK = {
    '0001' : 'Routing Information Packet',
    '0002' : 'Echo Protocol Packet',
    '0003' : 'Error Handling Packet',
    '0451' : 'NCP',                         # Netware Core Protocol - used by Novell Netware servers
    '0452' : 'SAP',                         # Service Advertising Protocol
    '0453' : 'RIP',                         # Routing Information Protocol
    '0455' : 'NetBIOS',
    '0456' : 'Diagnostic Packet',
    '0457' : 'Serialization Packet',        # used for NCP as well
    '4003' : 'Novell Netware Client',       # Used by Novell Netware Client
    '8060' : 'IPX',
    '9091' : 'TCP',                         # TCP over IPXF
    '9092' : 'UDP',                         # UDP over IPXF
    '9093' : 'IPXF',                        # IPX Fragmentation Protocol
}


class IPX(Internet):
    """This class implements Internetwork Packet Exchange.

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Internet`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance
        * src -- Info, source IPX address
        * dst -- Info, destination IPX address

    Methods:
        * read_ipx -- read Internetwork Packet Exchange

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor
        * _read_ipx_address -- read IPX address field

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of corresponding procotol."""
        return 'Internetwork Packet Exchange'

    @property
    def length(self):
        """Header length of corresponding protocol."""
        return 30

    @property
    def protocol(self):
        """Name of next layer protocol."""
        return self._info.type

    @property
    def src(self):
        """Source IPX address."""
        return self._info.src.addr

    @property
    def dst(self):
        """Destination IPX address."""
        return self._info.dst.addr

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_ipx(self, length):
        """Read Internetwork Packet Exchange.

        Structure of IPX header [RFC 1132]:
            Octets      Bits        Name                    Discription
              0           0     ipx.cksum               Checksum
              2          16     ipx.len                 Packet Length (header includes)
              4          32     ipx.count               Transport Control (hop count)
              5          40     ipx.type                Packet Type
              6          48     ipx.dst                 Destination Address
              18        144     ipx.src                 Source Address

        """
        if length is None:
            length = len(self)

        _csum = self._read_fileng(2)
        _tlen = self._read_unpack(2)
        _ctrl = self._read_unpack(1)
        _type = self._read_unpack(1)
        _dsta = self._read_ipx_address()
        _srca = self._read_ipx_address()

        ipx = dict(
            chksum = _csum,
            len = _tlen,
            count = _ctrl,
            type = TYPE.get(_type),
            dst = _dsta,
            src = _srca,
        )

        proto = ipx['type']
        length = ipx['len'] - 30
        ipx['packet'] = self._read_packet(header=30, payload=length)

        return self._read_next_layer(ipx, proto, length)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, **kwargs):
        self._file = _file
        self._info = Info(self.read_ipx(length))

    def __length_hint__(self):
        return 30

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _read_ipx_address(self):
        """Read IPX address field.

        Structure of IPX address:
            Octets      Bits        Name                    Discription
              0           0     ipx.addr.network        Network Number
              4          32     ipx.addr.node           Node Number
              10         80     ipx.addr.socket         Socket Number

        """
        # Adress Number
        _byte = self._read_fileng(4)
        _ntwk = ':'.join(textwrap.wrap(_byte.hex(), 2))

        # Node Number (MAC)
        _byte = self._read_fileng(6)
        _node = ':'.join(textwrap.wrap(_byte.hex(), 2))
        _maca = '-'.join(textwrap.wrap(_byte.hex(), 2))

        # Socket Number
        _sock = self._read_fileng(2)

        # Whole Address
        _list = [_ntwk, _node, _sock.hex()]
        _addr = ':'.join(_list)

        addr = dict(
            network = _ntwk,
            node = _maca,
            socket = SOCK.get(_sock.hex()) or _sock,
            addr = _addr,
        )

        return addr



# -*- coding: utf-8 -*-
"""stream pcap file extractor

`jspcap` is an independent open source library, using only
[`jsformat`](https://github.com/JarryShaw/jsformat) as
its formatted output dumper.

    There is a project called
    [`jspcapy](https://github.com/JarryShaw/jspcapy)
    works on `jspcap`, which is a command line tool for
    PCAP extraction.

Unlike popular PCAP file extractors, such as `Scapy`,
`dkpt`, `pyshark`, and etc, `jspcap` uses streaming
strategy to read input files. That is to read frame by
frame, decrease occupation on memory, as well as enhance
efficiency in some way.

In `jspcap`, all files can be described as following eight
different sections.

 - Foundation (`jspcap.foundation`)
    synthesise file I/O and protocol analysis, coordinate
    information exchange in all network layers

 - Interface (`jspcap.interface`)
    user interface for the `jspcap` library, which
    standardise and simplify the usage of this
    library

 - Reassembly (`jspcap.reassembly`)
    base on algorithms described in
    [`RFC 815`](https://tools.ietf.org/html/rfc815>),
    implement datagram reassembly of IP and TCP packets

 - IPSuite (`jspcap.ipsuite`)
    collection of constructors for Internet Protocol Suite

 - Protocols (`jspcap.protocols`)
    collection of all protocol family, with detailed
    implementation and methods

 - Utilities (`jspcap.utilities`)
    collection of utility functions and classes

 - CoreKit (`jspcap.corekit`)
    core utilities for `jspcap` implementation

 - DumpKit (`jspcap.dumpkit`)
    dump utilities for `jspcap` implementation

"""
# Interface
from jspcap.interface import *

# Protocols
from jspcap.protocols.raw import Raw
from jspcap.protocols.link.arp import ARP
from jspcap.protocols.link.ethernet import Ethernet
from jspcap.protocols.link.l2tp import L2TP
from jspcap.protocols.link.ospf import OSPF
from jspcap.protocols.link.rarp import RARP
from jspcap.protocols.link.vlan import VLAN
from jspcap.protocols.internet.ah import AH
from jspcap.protocols.internet.hip import HIP
from jspcap.protocols.internet.hopopt import HOPOPT
from jspcap.protocols.internet.ip import IP
from jspcap.protocols.internet.ipsec import IPsec
from jspcap.protocols.internet.ipv4 import IPv4
from jspcap.protocols.internet.ipv6 import IPv6
from jspcap.protocols.internet.ipv6_frag import IPv6_Frag
from jspcap.protocols.internet.ipv6_opts import IPv6_Opts
from jspcap.protocols.internet.ipv6_route import IPv6_Route
from jspcap.protocols.internet.ipx import IPX
from jspcap.protocols.internet.mh import MH
from jspcap.protocols.transport.tcp import TCP
from jspcap.protocols.transport.udp import UDP
from jspcap.protocols.application.http import HTTP


__all__ = [
    'TREE', 'JSON', 'PLIST', 'PCAP',                        # Macros
    'extract', 'analyse', 'reassemble', 'trace',            # Functions
    'Raw',                                                  # Raw Packet
    'ARP', 'Ethernet', 'L2TP', 'OSPF', 'RARP', 'VLAN',      # Link Layer
    'AH', 'IP', 'IPsec', 'IPv4', 'IPv6', 'IPX',             # Internet Layer
    'HIP', 'HOPOPT', 'IPv6_Frag', 'IPv6_Opts', 'IPv6_Route', 'MH',
                                                            # IPv6 Extension Header
    'TCP', 'UDP',                                           # Transport Layer
    'HTTP',                                                 # Application Layer
]



# -*- coding: utf-8 -*-
"""global header

`jspcap.protocols.pcap.header` contains `Header` only,
which implements extractor for global headers of PCAP,
whose structure is described as below.

typedef struct pcap_hdr_s {
    guint32 magic_number;   /* magic number */
    guint16 version_major;  /* major version number */
    guint16 version_minor;  /* minor version number */
    gint32  thiszone;       /* GMT to local correction */
    guint32 sigfigs;        /* accuracy of timestamps */
    guint32 snaplen;        /* max length of captured packets, in octets */
    guint32 network;        /* data link type */
} pcap_hdr_t;

"""
from jspcap.corekit.infoclass import Info
from jspcap.corekit.version import VersionInfo
from jspcap.protocols.protocol import Protocol
from jspcap.protocols.link.link import LINKTYPE
from jspcap.utilities.exceptions import FileError, UnsupportedCall
from jspcap.utilities.validations import int_check


__all__ = ['Header']


class Header(Protocol):
    """PCAP file global header extractor.

    Properties:
        * name -- str, name of corresponding protocol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding protocol
        * length -- int, header length of global header, i.e. 24
        * version -- VersionInfo, version infomation of input PCAP file
        * protocol -- str, data link type

    Methods:
        * index -- call `ProtoChain.index`
        * read_header -- read global header of PCAP file

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of corresponding protocol."""
        return 'Global Header'

    @property
    def length(self):
        """Header length of corresponding protocol."""
        return 24

    @property
    def version(self):
        """Version infomation of input PCAP file."""
        return VersionInfo(self._info.version_major, self._info.version_minor)

    @property
    def protocol(self):
        """Data link type."""
        return self._info.network

    @property
    def protochain(self):
        """NotImplemented"""
        raise UnsupportedCall("'Header' object has no attribute 'protochain'")

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_header(self):
        """Read global header of PCAP file.

        Structure of global header (C):
            typedef struct pcap_hdr_s {
                guint32 magic_number;   /* magic number */
                guint16 version_major;  /* major version number */
                guint16 version_minor;  /* minor version number */
                gint32  thiszone;       /* GMT to local correction */
                guint32 sigfigs;        /* accuracy of timestamps */
                guint32 snaplen;        /* max length of captured packets, in octets */
                guint32 network;        /* data link type */
            } pcap_hdr_t;

        """
        _temp = self._read_fileng(4)
        if _temp != b'\xd4\xc3\xb2\xa1':
            raise FileError('Unknown file format.')

        _magn = _temp
        _vmaj = self._read_unpack(2, lilendian=True)
        _vmin = self._read_unpack(2, lilendian=True)
        _zone = self._read_unpack(4, lilendian=True, signed=True)
        _acts = self._read_unpack(4, lilendian=True)
        _slen = self._read_unpack(4, lilendian=True)
        _type = self._read_protos(4)

        header = dict(
            magic_number = _magn,
            version_major = _vmaj,
            version_minor = _vmin,
            thiszone = _zone,
            sigfigs = _acts,
            snaplen = _slen,
            network = _type,
        )

        return header

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, file, **kwrags):
        self._file = file
        self._info = Info(self.read_header())

    def __len__(self):
        return 24

    def __length_hint__(self):
        return 24

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _read_protos(self, size):
        """Read next layer protocol type.

        Positional arguments:
            * size  -- int, buffer size

        Returns:
            * str -- link layer protocol name

        """
        _byte = self._read_unpack(4, lilendian=True)
        _prot = LINKTYPE.get(_byte)
        return _prot

    def _decode_next_layer(self, dict_, proto=None, length=None):
        """Deprecated."""
        raise UnsupportedCall(f"'{self.__class__.__name__}' object has no attribute '_decode_next_layer'")

    def _import_next_layer(self, proto, length):
        """Deprecated."""
        raise UnsupportedCall(f"'{self.__class__.__name__}' object has no attribute '_import_next_layer'")



# -*- coding: utf-8 -*-
"""

"""
from jspcap.ipsuite.protocol import Protocol
from jspcap.protocols.link.link import LINKTYPE
from jspcap.utilities.exceptions import ProtocolNotImplemented


__all__ = ['Header']


class Header(Protocol):
    """PCAP global header constructor.

    Properties:
        * name -- str, name of corresponding protocol
        * info -- Info, info dict of current instance
        * data -- bytes, binary packet data if current instance
        * alias -- str, acronym of corresponding protocol

    Methods:
        * index -- return first index of value from a dict
        * pack -- pack integers to bytes
        * update -- update packet data

    """
    ##########################################################################
    # Methods.
    ##########################################################################

    def update(self, **kwargs):
        """Update packet data."""
        # update dict
        self.__dict__.update(kwargs)

        # fetch values
        version_major = self.__dict__.pop('version_major', 2)   # major version number
        version_minor = self.__dict__.pop('version_minor', 4)   # minor version number
        thiszone = self.__dict__.pop('thiszone', 0)             # GMT to local correction
        sigfigs = self.__dict__.pop('sigfigs', 0)               # accuracy of timestamps
        snaplen = self.__dict__.pop('snaplen', 262144)          # max length of captured packets, in octets
        network = self.__dict__.pop('network', 'Ethernet')      # data link type

        # update packet
        data = b'\xd4\xc3\xb2\xa1'
        data += self.pack(version_major, size=2, lilendian=True)
        data += self.pack(version_minor, size=2, lilendian=True)
        data += self.pack(thiszone, size=4, lilendian=True)
        data += self.pack(sigfigs, size=4, lilendian=True)
        data += self.pack(snaplen, size=4, lilendian=True)
        data += self.index(LINKTYPE, network, pack=True, lilendian=True)

        # update data
        self.__data__ = data



# -*- coding: utf-8 -*-
"""root transport layer protocol

`jspcap.protocols.transport.transport` contains both
`TP_PROTO` and `Transport`. The former is a dictionary
of transport layer protocol numbers, registered in IANA.
And the latter is a base class for transport layer
protocols, eg. TCP and UDP.

"""
import io

from jspcap.protocols.protocol import Protocol
from jspcap.utilities.decorators import beholder_ng

###############################################################################
# from jspcap.fundation.analysis import analyse
###############################################################################


__all__ = ['Transport', 'TP_PROTO']


# Transport Layer Protocol Numbers
TP_PROTO = {
    # Internet Layer
    1:  'ICMP',     # Internet Control Message Protocol
    2:  'IGMP',     # Internet Group Management Protocol
    4:  'IP',       # IP in IP (encapsulation)
   41:  'IPv6',     # IPv6 Encapsulation
   58:  'ICMPv6',   # ICMP for IPv6

    # IPv6 Extension Header Types
    0:  'HOPOPT',       # IPv6 Hop-by-Hop Option
   43:  'IPv6-Route',   # Routing Header for IPv6
   44:  'IPv6-Frag',    # Fragment Header for IPv6
   50:  'ESP',          # Encapsulating Security Payload
   51:  'AH',           # Authentication Header
   59:  'IPv6-NoNxt',   # No Next Header for IPv6
   60:  'IPv6-Opts',    # Destination Options for IPv6 (before routing / upper-layer header)
  135:  'Mobility',     # Mobility Extension Header for IPv6 (currently without upper-layer header)
  139:  'HIP',          # Host Identity Protocol
  140:  'Shim6',        # Site Multihoming by IPv6 Intermediation

   # Transport Layer
    6:  'TCP',      # Transmission Control Protocol
   17:  'UDP',      # User Datagram Protocol
   89:  'OSPF',     # Open Shortest Path First
  132:  'SCTP',     # Stream Control Transmission Protocol
}


class Transport(Protocol):
    """Abstract base class for transport layer protocol family.

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * layer -- str, `Transport`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor

    """
    __layer__ = 'Transport'

    ##########################################################################
    # Properties.
    ##########################################################################

    # protocol layer
    @property
    def layer(self):
        """Protocol layer."""
        return self.__layer__

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _import_next_layer(self, proto, length):
        """Import next layer extractor.

        Positional arguments:
            * proto -- str, next layer protocol name
            * length -- int, valid (not padding) length

        Returns:
            * bool -- flag if extraction of next layer succeeded
            * Info -- info of next layer
            * ProtoChain -- protocol chain of next layer
            * str -- alias of next layer

        """
        from jspcap.foundation.analysis import Analysis
        if self._onerror:
            next_ = beholder_ng(Analysis.analyse)(io.BytesIO(self._read_fileng(length)), length, _termination=self._sigterm)
        else:
            next_ = Analysis.analyse(io.BytesIO(self._read_fileng(length)), length, _termination=self._sigterm)
        return True, next_.info, next_.protochain, next_.alias



# -*- coding: utf-8 -*-
"""ethernet protocol

`jspcap.protocols.link.ethernet` contains `Ethernet`
only, which implements extractor for Ethernet Protocol,
whose structure is described as below.

Octets      Bits        Name                    Discription
  0           0     eth.dst                 Destination MAC Address
  1           8     eth.src                 Source MAC Address
  2          16     eth.type                Protocol (Internet Layer)

"""
import textwrap

from jspcap.corekit.infoclass import Info
from jspcap.protocols.link.link import Link


__all__ = ['Ethernet']


class Ethernet(Link):
    """This class implements Ethernet Protocol.

    Properties:
        * name -- str, name of corresponding protocol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding protocol
        * layer -- str, `Link`
        * length -- int, header length of corresponding protocol
        * protocol -- str, next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance
        * src -- str, source mac address
        * dst -- str, destination mac address

    Methods:
        * read_ethernet -- read Ethernet Protocol

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor
        * _read_mac_addr -- read MAC address

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of current protocol."""
        return 'Ethernet Protocol'

    @property
    def length(self):
        """Header length of current protocol."""
        return 14

    @property
    def protocol(self):
        """Name of next layer protocol."""
        return self._info.type

    # source mac address
    @property
    def src(self):
        """Source mac address."""
        return self._info.src

    # destination mac address
    @property
    def dst(self):
        """Destination mac address."""
        return self._info.dst

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_ethernet(self, length):
        """Read Ethernet Protocol.

        Structure of Ethernet Protocol header [RFC 7042]:
            Octets      Bits        Name                    Discription
              0           0     eth.dst                 Destination MAC Address
              1           8     eth.src                 Source MAC Address
              2          16     eth.type                Protocol (Internet Layer)

        """
        if length is None:
            length = len(self)

        _dstm = self._read_mac_addr()
        _srcm = self._read_mac_addr()
        _type = self._read_protos(2)

        ethernet = dict(
            dst = _dstm,
            src = _srcm,
            type = _type,
        )

        length -= 14
        ethernet['packet'] = self._read_packet(header=14, payload=length)

        return self._decode_next_layer(ethernet, _type, length)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, **kwargs):
        self._file = _file
        self._info = Info(self.read_ethernet(length))

    def __length_hint__(self):
        return 14

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _read_mac_addr(self):
        """Read MAC address."""
        _byte = self._read_fileng(6)
        _addr = '-'.join(textwrap.wrap(_byte.hex(), 2))
        return _addr



# -*- coding: utf-8 -*-
"""dump utilities

`jspcap.dumpkit` is the collection of dumpers for
`jspcap` implementation, which is alike those described
in [`jsformat`](https://github.com/JarryShaw/jsformat).

"""
from jspcap.ipsuite.pcap.frame import Frame
from jspcap.ipsuite.pcap.header import Header


__all__ = ['PCAP']



class NotImplementedIO:
    """Unspecified output format."""
    @property
    def kind(self):
        return NotImplemented

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        pass


class PCAP:
    """PCAP file dumper."""
    @property
    def kind(self):
        return 'pcap'

    def __init__(self, filename, *, protocol):
        self._file = filename
        packet = Header(protocol=protocol).data
        with open(self._file, 'wb') as file:
            file.write(packet)

    def __call__(self, frame, **kwargs):
        packet = Frame(frame).data
        with open(self._file, 'ab') as file:
            file.write(packet)



# -*- coding: utf-8 -*-
"""root protocol

`jspcap.protocols.protocol` contains `Protocol` only,
which is an abstract base clss for all protocol family,
with pre-defined utility arguments and methods of specified
protocols.

"""
import abc
import copy
import io
import numbers
import os
import re
import struct
import textwrap

from jspcap.corekit.infoclass import Info
from jspcap.corekit.protochain import ProtoChain
from jspcap.utilities.decorators import beholder, seekset
from jspcap.utilities.exceptions import BoolError, BytesError, StructError
from jspcap.utilities.validations import bool_check, int_check

###############################################################################
# from jspcap.protocols.raw import Raw
###############################################################################


__all__ = ['Protocol']


# abstract base class utilities
ABCMeta = abc.ABCMeta
abstractmethod = abc.abstractmethod
abstractproperty = abc.abstractproperty


class Protocol:
    """Abstract base class for all protocol family.

    Properties:
        * name -- str, name of corresponding protocol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding protocol
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor

    """
    __layer__ = None
    __metaclass__ = ABCMeta

    ##########################################################################
    # Properties.
    ##########################################################################

    # name of current protocol
    @abstractproperty
    def name(self):
        """Name of current protocol."""
        pass

    # acronym of current protocol
    @property
    def alias(self):
        """Acronym of current protocol."""
        return self.__class__.__name__

    # info dict of current instance
    @property
    def info(self):
        """Info dict of current instance."""
        return self._info

    # header length of current protocol
    @abstractproperty
    def length(self):
        """Header length of current protocol."""
        pass

    # name of next layer protocol
    @property
    def protocol(self):
        """Name of next layer protocol."""
        try:
            return self._protos[1]
        except IndexError:
            return None

    # protocol chain of current instance
    @property
    def protochain(self):
        """Protocol chain of current instance."""
        return self._protos

    ##########################################################################
    # Data models.
    ##########################################################################

    # Not hashable
    __hash__ = None

    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        self._onerror = kwargs.pop('error', False)
        self._exlayer = kwargs.pop('layer', str())
        self._exproto = kwargs.pop('protocol', str())
        self._sigterm = self._check_termination()
        return self

    def __repr__(self):
        name = self.__class__.__name__
        repr_ = f"<class 'protocol.{name}'>"
        return repr_

    @seekset
    def __str__(self):
        str_ = ' '.join(textwrap.wrap(self._file.read().hex(), 2))
        return str_

    @seekset
    def __bytes__(self):
        bytes_ = self._file.read()
        return bytes_

    @seekset
    def __len__(self):
        """Total length of correspoding protocol."""
        return len(self._read_fileng())

    @abstractmethod
    def __length_hint__(self):
        pass

    def __iter__(self):
        file = copy.deepcopy(self._file)
        file.seek(os.SEEK_SET)
        return iter(file)

    # def __next__(self):
    #     next_ = self._file.read(1)
    #     if next_:
    #         return next_
    #     else:
    #         self._file.seek(os.SEEK_SET)
    #         raise StopIteration

    def __getitem__(self, key):
        return self._info[key]

    @classmethod
    def __index__(cls):
        return cls.__name__

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _read_protos(self, size):
        """Read next layer protocol type.

        Positional arguments:
            * size  -- int, buffer size

        Returns:
            * [upon success] str -- name of next layer protocol
            * [upon failure] None

        """
        return None

    def _read_fileng(self, *args, **kwargs):
        """Read file buffer."""
        return self._file.read(*args, **kwargs)

    def _read_unpack(self, size=1, *, signed=False, lilendian=False, quiet=False):
        """Read bytes and unpack for integers.

        Positional arguments:
            * size  -- int, buffer size (default is 1)

        Keyword arguments:
            * signed -- bool, signed flag (default is False)
                           <keyword> True / False
            * lilendian -- bool, little-endian flag (default is False)
                           <keyword> True / False
            * quiet -- bool, quiet (no exception) flag (default is False)
                           <keyword> True / False

        Returns:
            * int -- unpacked data upon success

        """
        endian = '<' if lilendian else '>'
        if size == 8:   kind = 'q' if signed else 'Q'   # unpack to 8-byte integer (long long)
        elif size == 4: kind = 'i' if signed else 'I'   # unpack to 4-byte integer (int / long)
        elif size == 2: kind = 'h' if signed else 'H'   # unpack to 2-byte integer (short)
        elif size == 1: kind = 'b' if signed else 'B'   # unpack to 1-byte integer (char)
        else:           kind = None                     # do not unpack

        if kind is None:
            mem = self._file.read(size)
            end = 'little' if lilendian else 'big'
            buf = int.from_bytes(mem, end, signed=signed)
        else:
            try:
                fmt = f'{endian}{kind}'
                mem = self._file.read(size)
                buf = struct.unpack(fmt, mem)[0]
            except struct.error:
                if quiet:
                    return None
                else:
                    raise StructError(f'{self.__class__.__name__}: unpack failed') from None
        return buf

    def _read_binary(self, size=1):
        """Read bytes and convert into binaries.

        Positional arguments:
            * size  -- int, buffer size (default is 1)

        Returns:
            * str -- binary bits (0/1)

        """
        bin_ = ''
        for _ in range(size):
            byte = self._file.read(1)
            bin_ += bin(ord(byte))[2:].zfill(8)
        return bin_

    @seekset
    def _read_packet(self, length=None, *, header=None, payload=None, discard=False):
        """Read raw packet data.

        Positional arguments:
            * length -- int, length of the packet

        Keyword arguments:
            * header -- int, length of the packet header
            * payload -- int, length of the packet payload
            * discard -- bool, flag if discard header data (False in default)

        Returns:
            * [if header omits] bytes -- whole packet data
            * [if discard set True] bytes -- packet body only
            * dict -- header and payload data
                |-- 'header' -- bytes, packet header
                |-- 'payload' -- bytes, packet payload

        """
        if header is not None:
            header = self._read_fileng(header)
            payload = self._read_fileng(*[payload])
            if discard:
                return payload
            return dict(header=header, payload=payload)
        return self._read_fileng(*[length])

    def _decode_next_layer(self, dict_, proto=None, length=None):
        """Decode next layer protocol.

        Positional arguments:
            * dict_ -- dict, info buffer
            * proto -- str, next layer protocol name
            * length -- int, valid (not padding) length

        Returns:
            * dict -- current protocol with next layer extracted

        """
        if self._onerror:
            flag, info, chain, alias = beholder(self._import_next_layer)(self, proto, length)
        else:
            flag, info, chain, alias = self._import_next_layer(proto, length)

        # make next layer protocol name
        if flag:
            if proto is None and chain:
                layer = chain.alias[0].lower()
                proto, chain = chain.tuple[0], None
            else:
                layer = str(alias or proto or 'Raw').lower()
        else:
            layer, proto = 'raw', 'Raw'

        # write info and protocol chain into dict
        dict_[layer] = info
        self._protos = ProtoChain(proto, chain, alias)
        return dict_

    def _import_next_layer(self, proto, length=None):
        """Import next layer extractor.

        Positional arguments:
            * proto -- str, next layer protocol name
            * length -- int, valid (not padding) length

        Returns:
            * bool -- flag if extraction of next layer succeeded
            * Info -- info of next layer
            * ProtoChain -- protocol chain of next layer
            * str -- alias of next layer

        """
        from jspcap.protocols.raw import Raw
        next_ = Raw(io.BytesIO(self._read_fileng(length)), length,
                    layer=self._exlayer, protocol=self._exproto)
        return True, next_.info, next_.protochain, next_.alias

    def _check_termination(self):
        index = self.__index__()
        pattern = '|'.join(index) if isinstance(index, tuple) else index
        iterable = self._exproto if isinstance(self._exproto, tuple) else (self._exproto,)
        match = filter(lambda string: re.fullmatch(pattern, string, re.IGNORECASE), iterable)
        return (len(list(match)) or re.fullmatch(self.__layer__ or '', self._exlayer, re.IGNORECASE))



# -*- coding: utf-8 -*-
"""core utilities

`jspcap.corekit` is the collection of core utilities for
`jspcap` implementation, including dict-like class `Info`,
tuple-like class `VersionInfo`, and protocol collection
class `ProtoChain`.

"""
from jspcap.corekit.infoclass import Info
from jspcap.corekit.protochain import ProtoChain
from jspcap.corekit.version import VersionInfo


__all__ = ['Info', 'ProtoChain', 'VersionInfo']



# -*- coding: utf-8 -*-
"""utility functions and classes

`jspcap.utilities` contains several useful functions and
classes which are fundations of `jspcap`, including
decorater function `seekset` and `beholder`, and several
user-refined exceptions and validations.

"""
from jspcap.utilities.decorators import *
from jspcap.utilities.exceptions import *
from jspcap.utilities.validations import *
from jspcap.utilities.warnings import *


__all__ = ['seekset_ng', 'beholder_ng']



# -*- coding: utf-8 -*-
"""host identity protocol

`jspcap.protocols.internet.hip` contains `HIP`
only, which implements extractor for Host Identity
Protocol (HIP), whose structure is described as below.

 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
| Next Header   | Header Length |0| Packet Type |Version| RES.|1|
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|          Checksum             |           Controls            |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                Sender's Host Identity Tag (HIT)               |
|                                                               |
|                                                               |
|                                                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|               Receiver's Host Identity Tag (HIT)              |
|                                                               |
|                                                               |
|                                                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                                                               |
/                        HIP Parameters                         /
/                                                               /
|                                                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

"""
# TODO: Implements extractor of all HIP parameters.
from jspcap.corekit.infoclass import Info
from jspcap.protocols.protocol import Protocol
from jspcap.utilities.exceptions import ProtocolError


__all__ = ['HIP']


# HIP Packet Types
_HIP_TYPES = {
    0 : 'Reserved',     # [RFC 7401]
    1 : 'I1',           # [RFC 7401] the HIP Initiator Packet
    2 : 'R1',           # [RFC 7401] the HIP Responder Packet
    3 : 'I2',           # [RFC 7401] the Second HIP Initiator Packet
    4 : 'R2',           # [RFC 7401] the Second HIP Responder Packet
   16 : 'UPDATE',       # [RFC 7401] the HIP Update Packet
   17 : 'NOTIFY',       # [RFC 7401] the HIP Notify Packet
   18 : 'CLOSE',        # [RFC 7401] the HIP Association Closing Packet
   19 : 'CLOSE_ACK',    # [RFC 7401] the HIP Closing Acknowledgment Packet
   20 : 'HDRR',         # [RFC 6537] HIP Distributed Hash Table Resource Record
   32 : 'HIP_DATA',     # [RFC 6078]
}


class HIP(Protocol):
    """This class implements Host Identity Protocol.

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Internet`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Methods:
        * read_hip -- read Host Identity Protocol (HIP)

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of current protocol."""
        return 'Host Identity Protocol'

    @property
    def alias(self):
        """Acronym of corresponding procotol."""
        return f'HIPv{self._info.version}'

    @property
    def length(self):
        """Header length of current protocol."""
        return self._info.length

    @property
    def protocol(self):
        """Name of next layer protocol."""
        return self._info.next

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_hip(self, length, extension):
        """Read Host Identity Protocol.

        Structure of HIP header [RFC 7401]:
             0                   1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            | Next Header   | Header Length |0| Packet Type |Version| RES.|1|
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |          Checksum             |           Controls            |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                Sender's Host Identity Tag (HIT)               |
            |                                                               |
            |                                                               |
            |                                                               |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |               Receiver's Host Identity Tag (HIT)              |
            |                                                               |
            |                                                               |
            |                                                               |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                                                               |
            /                        HIP Parameters                         /
            /                                                               /
            |                                                               |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

            Octets      Bits        Name                    Discription
              0           0     hip.next                Next Header
              1           8     hip.length              Header Length
              2          16     -                       Reserved (0)
              2          17     hip.type                Packet Type
              3          24     hip.version             Version
              3          28     -                       Reserved
              3          31     -                       Reserved (1)
              4          32     hip.chksum              Checksum
              6          48     hip.control             Controls
              8          64     hip.shit                Sender's Host Identity Tag
              24        192     hip.rhit                Receiver's Host Identity Tag
              40        320     hip.parameters          HIP Parameters

        """
        if length is None:
            length = len(self)

        _next = self._read_protos(1)
        _hlen = self._read_unpack(1)
        _type = self._read_binary(1)
        if _type[0] != '0':
            raise ProtocolError('HIP: invalid format')
        _vers = self._read_binary(1)
        if _vers[7] != '1':
            raise ProtocolError('HIP: invalid format')
        _csum = self._read_fileng(2)
        _ctrl = self._read_binary(2)
        _shit = self._read_unpack(16)
        _rhit = self._read_unpack(16)
        _para = self._read_fileng(_hlen - 38)

        hip = dict(
            next = _next,
            length = _hlen + 1,
            type = _HIP_TYPES.get(int(_type[1:], base=2), 'Unassigned'),
            version = int(_vers[:4], base=2),
            chksum = _csum,
            control = dict(
                anonymous = True if int(_ctrl[15], base=2) else False,
            ),
            shit = _shit,
            rhit = _rhit,
            parameters = _para,
        )

        length -= hip['length']
        hip['packet'] = self._read_packet(header=hip['length'], payload=length)

        if extension:
            self._protos = None
            return hip
        return self._decode_next_layer(hip, _next, length)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, *, extension=False, **kwargs):
        self._file = _file
        self._info = Info(self.read_hip(length, extension))

    def __length_hint__(self):
        return 40



# -*- coding: utf-8 -*-
"""fragment header for IPv6

`jspcap.protocols.internet.ipv6_frag` contains `IPv6_Frag`
only, which implements extractor for Fragment Header for
IPv6 (IPv6-Frag), whose structure is described as below.

+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|  Next Header  |   Reserved    |      Fragment Offset    |Res|M|
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                         Identification                        |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

"""
from jspcap.corekit.infoclass import Info
from jspcap.protocols.protocol import Protocol


__all__ = ['IPv6_Frag']


class IPv6_Frag(Protocol):
    """This class implements Fragment Header for IPv6.

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Internet`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Methods:
        * read_ipv6_frag -- read Fragment Header for IPv6 (IPv6-Frag)

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of current protocol."""
        return 'Fragment Header for IPv6'

    @property
    def alias(self):
        """Acronym of corresponding procotol."""
        return 'IPv6-Frag'

    @property
    def length(self):
        """Header length of current protocol."""
        return self._info.length

    @property
    def protocol(self):
        """Name of next layer protocol."""
        return self._info.next

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_ipv6_frag(self, length, extension):
        """Read Fragment Header for IPv6.

        Structure of IPv6-Frag header [RFC 8200]:
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |  Next Header  |   Reserved    |      Fragment Offset    |Res|M|
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                         Identification                        |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

            Octets      Bits        Name                    Discription
              0           0     frag.next               Next Header
              1           8     -                       Reserved
              2          16     frag.offset             Fragment Offset
              3          29     -                       Reserved
              3          31     frag.mf                 More Flag
              4          32     frag.id                 Identification

        """
        if length is None:
            length = len(self)

        _next = self._read_protos(1)
        _temp = self._read_fileng(1)
        _offm = self._read_binary(2)
        _ipid = self._read_unpack(4)

        ipv6_frag = dict(
            next = _next,
            offset = int(_offm[:13], base=2),
            mf = True if int(_offm[15], base=2) else False,
            id = _ipid,
        )

        if length is not None:
            length -= 8
        ipv6_frag['packet'] = self._read_packet(header=8, payload=length)

        if extension:
            self._protos = None
            return ipv6_frag
        return self._decode_next_layer(ipv6_frag, _next, length)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, *, extension=False, **kwargs):
        self._file = _file
        self._info = Info(self.read_ipv6_frag(length, extension))

    def __length_hint__(self):
        return 8
